"""
Provide tool functions for working with Cluster data
"""

import json
from datetime import datetime
from typing import Dict, List, Optional, cast

from loguru import logger

from jobbergate_cli.config import settings
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import ClusterCacheData, JobbergateContext
from jobbergate_cli.text_tools import conjoin


def pull_client_ids_from_api(ctx: JobbergateContext) -> List[str]:
    assert ctx.client is not None

    response_data = cast(
        Dict,
        make_request(
            ctx.client,
            "/cluster/graphql/query",
            "POST",
            expected_status=200,
            abort_message="There was a problem retrieving registered clusters from the Cluster API",
            abort_subject="COULD NOT RETRIEVE CLUSTERS",
            support=True,
            json=dict(
                query="query {cluster{clientId}}",
                variables=dict(),
            ),
        ),
    )

    try:
        client_ids = [e["clientId"] for e in response_data["data"]["cluster"]]
    except Exception as err:
        raise Abort(
            "Couldn't unpack cluster names from Cluster API response",
            subject="COULD NOT RETRIEVE CLUSTERS",
            support=True,
            original_error=err,
            log_message=f"Failed to unpack data from cluster-api: {response_data}",
        )
    return client_ids


def save_clusters_to_cache(client_ids: List[str]):

    # Make static type checkers happy
    assert settings.JOBBERGATE_CLUSTER_LIST_PATH is not None

    cache_data = ClusterCacheData(
        updated_at=datetime.utcnow(),
        client_ids=client_ids,
    )

    logger.debug(f"Caching cluster info at {settings.JOBBERGATE_CLUSTER_LIST_PATH}")
    settings.JOBBERGATE_CLUSTER_LIST_PATH.write_text(cache_data.json())


def load_clusters_from_cache() -> Optional[List[str]]:

    # Make static type checkers happy
    assert settings.JOBBERGATE_CLUSTER_LIST_PATH is not None

    try:
        cache_data = ClusterCacheData(**json.loads(settings.JOBBERGATE_CLUSTER_LIST_PATH.read_text()))
    except Exception as err:
        logger.warning(f"Couldn't load cluster data from cache: {err}")
        return None

    if datetime.utcnow().timestamp() - cache_data.updated_at.timestamp() > settings.JOBBERGATE_CLUSTER_CACHE_LIFETIME:
        logger.warning("Cached cluster data is expired")
        return None

    return cache_data.client_ids


def get_client_ids(ctx: JobbergateContext) -> List[str]:
    assert ctx.client is not None

    client_ids = load_clusters_from_cache()
    if client_ids is None:
        client_ids = pull_client_ids_from_api(ctx)
        save_clusters_to_cache(client_ids)

    return client_ids


def validate_client_id(ctx: JobbergateContext, client_id: str):
    client_ids = get_client_ids(ctx)
    Abort.require_condition(
        client_id in client_ids,
        conjoin(
            """
            The supplied cluster name was not found in the list of available clusters.

            Please select one of:
            """,
            *client_ids,
        ),
        raise_kwargs=dict(
            subject="Invalid cluster name",
            support=True,
        ),
    )



"""
Provide a ``typer`` app that can interact with Application data in a cruddy manner.
"""

import pathlib
from typing import Any, Dict, Optional, cast

import typer

from jobbergate_cli.constants import OV_CONTACT, SortOrder
from jobbergate_cli.exceptions import handle_abort
from jobbergate_cli.render import StyleMapper, render_list_results, render_single_result, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import JobbergateContext, ListResponseEnvelope
from jobbergate_cli.subapps.applications.tools import fetch_application_data, load_default_config, upload_application


# TODO: move hidden field logic to the API
HIDDEN_FIELDS = [
    "application_config",
    "application_source_file",
    "application_templates",
    "created_at",
    "updated_at",
]

ID_NOTE = """

    This id represents the primary key of the application in the database. It
    will always be a unique integer and is automatically generated by the server
    when an Application is created. All applications receive an id, so it may
    be used to target a specific instance of an application whether or not it
    is provided with a human-friendly "identifier".
"""


IDENTIFIER_NOTE = """

    The identifier allows the user to access commonly used applications with a
    friendly name that is easy to remember. Identifiers should only be used
    for applications that are frequently used or should be easy to find in the list.
    An identifier may be added, removed, or changed on an existing application.
"""


style_mapper = StyleMapper(
    id="green",
    application_name="cyan",
    application_identifier="magenta",
)


app = typer.Typer(help="Commands to interact with applications")


@app.command("list")
@handle_abort
def list_all(
    ctx: typer.Context,
    show_all: bool = typer.Option(False, "--all", help="Show all applications, even the ones without identifier"),
    user_only: bool = typer.Option(False, "--user", help="Show only applications owned by the current user"),
    search: Optional[str] = typer.Option(None, help="Apply a search term to results"),
    sort_order: SortOrder = typer.Option(SortOrder.UNSORTED, help="Specify sort order"),
    sort_field: Optional[str] = typer.Option(None, help="The field by which results should be sorted"),
):
    """
    Show available applications
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx is not None
    assert jg_ctx.client is not None

    params: Dict[str, Any] = dict(
        all=show_all,
        user=user_only,
    )
    if search is not None:
        params["search"] = search
    if sort_order is not SortOrder.UNSORTED:
        params["sort_ascending"] = SortOrder is SortOrder.ASCENDING
    if sort_field is not None:
        params["sort_field"] = sort_field

    envelope = cast(
        ListResponseEnvelope,
        make_request(
            jg_ctx.client,
            "/jobbergate/applications",
            "GET",
            expected_status=200,
            abort_message="Couldn't retrieve applications list from API",
            support=True,
            response_model_cls=ListResponseEnvelope,
            params=params,
        ),
    )
    render_list_results(
        jg_ctx,
        envelope,
        title="Applications List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
    )


@app.command()
@handle_abort
def get_one(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        help=f"The specific id of the application. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
):
    """
    Get a single application by id or identifier
    """
    jg_ctx: JobbergateContext = ctx.obj
    result = fetch_application_data(jg_ctx, id=id, identifier=identifier)
    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Application",
    )


@app.command()
@handle_abort
def create(
    ctx: typer.Context,
    name: str = typer.Option(
        ...,
        help="The name of the application to create",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
    application_path: pathlib.Path = typer.Option(
        ...,
        help="The path to the directory where the application files are located",
    ),
    application_desc: Optional[str] = typer.Option(
        None,
        help="A helpful description of the application",
    ),
):
    """
    Create a new application.
    """
    req_data = load_default_config()
    req_data["application_name"] = name
    if identifier:
        req_data["application_identifier"] = identifier

    if application_desc:
        req_data["application_description"] = application_desc

    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    result = cast(
        Dict[str, Any],
        make_request(
            jg_ctx.client,
            "/jobbergate/applications",
            "POST",
            expected_status=201,
            abort_message="Request to create application was not accepted by the API",
            support=True,
            json=req_data,
        ),
    )
    application_id = result["id"]

    successful_upload = upload_application(jg_ctx, application_path, application_id)
    if not successful_upload:
        terminal_message(
            f"""
            The application files could not be uploaded.

            Try running the `update` command including the application path to re-upload.

            [yellow]If the problem persists, please contact [bold]{OV_CONTACT}[/bold]
            for support and trouble-shooting[/yellow]
            """,
            subject="File upload failed",
            color="yellow",
        )
    else:
        result["application_uploaded"] = True

    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Created Application",
    )


@app.command()
@handle_abort
def update(
    ctx: typer.Context,
    id: int = typer.Option(
        ...,
        help=f"The specific id of the application to update. {ID_NOTE}",
    ),
    application_path: Optional[pathlib.Path] = typer.Option(
        None,
        help="The path to the directory where the application files are located",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help="Optional new application identifier to be set",
    ),
    application_desc: Optional[str] = typer.Option(
        None,
        help="Optional new application description to be set",
    ),
):
    """
    Update an existing application.
    """
    req_data = dict()

    if identifier:
        req_data["application_identifier"] = identifier

    if application_desc:
        req_data["application_description"] = application_desc

    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    result = cast(
        Dict[str, Any],
        make_request(
            jg_ctx.client,
            f"/jobbergate/applications/{id}",
            "PUT",
            expected_status=200,
            abort_message="Request to update application was not accepted by the API",
            support=True,
            json=req_data,
        ),
    )

    if application_path is not None:
        successful_upload = upload_application(jg_ctx, application_path, id)
        if not successful_upload:
            terminal_message(
                f"""
                The application files could not be uploaded.

                [yellow]If the problem persists, please contact [bold]{OV_CONTACT}[/bold]
                for support and trouble-shooting[/yellow]
                """,
                subject="File upload failed",
                color="yellow",
            )
        else:
            result["application_uploaded"] = True

    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Updated Application",
    )


@app.command()
@handle_abort
def delete(
    ctx: typer.Context,
    id: int = typer.Option(
        ...,
        help=f"the specific id of the application to delete. {ID_NOTE}",
    ),
):
    """
    Delete an existing application.
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    # Delete the upload. The API will also remove the application data files
    make_request(
        jg_ctx.client,
        f"/jobbergate/applications/{id}",
        "DELETE",
        expected_status=204,
        expect_response=False,
        abort_message="Request to delete application was not accepted by the API",
        support=True,
    )
    terminal_message(
        """
        The application was successfully deleted.
        """,
        subject="Application delete succeeded",
    )



"""
Jobbergate command-line interface and app library
"""

import importlib
import warnings

# This is a backport of PEP-562 for python3.6
from pep562 import pep562

from jobbergate_cli.config import settings


def __getattr__(name: str):
    """
    Overload module attribute lookup to warn if 'appform' is being imported because it is deprecated.
    """
    if name == "appform" and settings.JOBBERGATE_COMPATIBILITY_MODE:
        warnings.warn(
            "appform is deprecated.  Import from 'jobbergate_cli.subapps.applications.questions' instead",
            DeprecationWarning,
        )
        return importlib.import_module("jobbergate_cli.subapps.applications.questions", __name__)

    else:
        raise AttributeError(f"module {__name__} has no attribute {name}")


pep562(__name__)



"""
Provide tool functions for working with Application data.
"""

import contextlib
import copy
import io
import pathlib
from typing import Any, Dict, Optional, Tuple, cast

import yaml
from loguru import logger

from jobbergate_cli.constants import JOBBERGATE_APPLICATION_CONFIG, JOBBERGATE_APPLICATION_SUPPORTED_FILES
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import ApplicationResponse, JobbergateApplicationConfig, JobbergateContext
from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase
from jobbergate_cli.subapps.applications.questions import gather_param_values


def load_default_config() -> Dict[str, Any]:
    """
    Load the default config for an application.
    """
    return copy.deepcopy(JOBBERGATE_APPLICATION_CONFIG)


def fetch_application_data(
    jg_ctx: JobbergateContext,
    id: Optional[int] = None,
    identifier: Optional[str] = None,
) -> ApplicationResponse:
    """
    Retrieve an application from the API by ``id`` or ``identifier``.

    :param: jg_ctx:     The JobbergateContext. Needed to access the Httpx client with which to make API calls
    :param: id:         The id of the application to fetch
    :param: identifier: If supplied, look for an application instance with the provided identifier
    :returns: An instance of ApplicationResponse containing the application data
    """
    identification: Any = id
    if id is None and identifier is None:
        raise Abort(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif id is not None and identifier is not None:
        raise Abort(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif identifier is not None:
        identification = identifier

    # Make static type checkers happy
    assert jg_ctx.client is not None

    stub = f"id={id}" if id is not None else f"identifier={identifier}"
    return cast(
        ApplicationResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/applications/{identification}",
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve application {stub} from API",
            response_model_cls=ApplicationResponse,
            support=True,
        ),
    )


def load_application_data(
    app_data: ApplicationResponse,
) -> Tuple[JobbergateApplicationConfig, JobbergateApplicationBase]:
    """
    Validates and loads the data for an application returned from the API's applications GET endpoint.

    :param: app_data: A dictionary containing the application data
    :returns: A tuple containing the application config and the application module
    """
    if not app_data.application_config:

        raise Abort(
            f"Fail to retrieve the application config file for id={app_data.id}",
            subject="Application config is missing",
            log_message="Application config is missing",
        )

    try:
        app_config = load_application_config_from_source(app_data.application_config)
    except Exception as err:
        print("ERR: ", err)
        raise Abort(
            "The application config fetched from the API is not valid",
            subject="Invalid application config",
            support=True,
            log_message="Invalid application config",
            original_error=err,
        )

    if not app_data.application_source_file:

        raise Abort(
            f"Fail to retrieve the application source file for id={app_data.id}",
            subject="Application source file is missing",
            log_message="Application source file is missing",
        )

    try:
        app_module = load_application_from_source(app_data.application_source_file, app_config)
    except Exception as err:
        raise Abort(
            "The application source fetched from the API is not valid",
            subject="Invalid application module",
            support=True,
            log_message="Invalid application module",
            original_error=err,
        )

    return (app_config, app_module)


@contextlib.contextmanager
def get_upload_files(application_path: pathlib.Path):
    """
    Context manager to build the ``files`` parameter.

    Open the supplied file(s) and build a ``files`` param appropriate for using
    multi-part file uploads with the client.
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    with contextlib.ExitStack() as stack:
        yield [
            (
                "upload_files",
                (path.name, stack.enter_context(io.open(path, mode="r", newline="")), "text/plain"),
            )
            for path in application_path.rglob("*")
            if path.is_file() and path.suffix in JOBBERGATE_APPLICATION_SUPPORTED_FILES
        ]


def upload_application(
    jg_ctx: JobbergateContext,
    application_path: pathlib.Path,
    application_id: int,
) -> bool:
    """
    Upload an application given an application path and the application id.

    :param: jg_ctx:           The JobbergateContext. Needed to access the Httpx client with which to make API calls
    :param: application_path: The directory where the application files to upload may be found
    :param: application_id:   The id of the application for which to upload  data
    :returns: True if the upload was successful; False otherwise
    """

    # Make static type checkers happy
    assert jg_ctx.client is not None

    with get_upload_files(pathlib.Path(application_path)) as upload_files:
        logger.debug(
            f"Preparing to upload {len(upload_files)} application files from {application_path}",
        )
        response_code = cast(
            int,
            make_request(
                jg_ctx.client,
                f"/jobbergate/applications/{application_id}/upload",
                "POST",
                expect_response=False,
                abort_message="Request to upload application files was not accepted by the API",
                support=True,
                files=upload_files,
            ),
        )
        return response_code == 201


def load_application_config_from_source(config_source: str) -> JobbergateApplicationConfig:
    """
    Load the JobbergateApplicationConfig from a text string containing the config as YAML.

    :param: config_source: The YAML containing the config
    :returns: A JobbergateApplicationConfig instance with the config values
    """
    config_data = yaml.safe_load(config_source)
    config = JobbergateApplicationConfig(**config_data)
    return config


def load_application_from_source(app_source: str, app_config: JobbergateApplicationConfig) -> JobbergateApplicationBase:
    """
    Load the JobbergateApplication class from a text string containing the source file.

    Creates the module in a temporary file and imports it with importlib.

    Adapted from: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly

    :param: app_source: The JobbergateApplication source code to load
    :param: app_config: The JobbergateApplicationConfig needed to instantiate the JobbergateApplication
    """
    app_locals: Dict[str, Any] = dict()
    exec(app_source, app_locals, app_locals)
    jobbergate_application_class = app_locals["JobbergateApplication"]
    application = jobbergate_application_class(app_config.dict())
    return application


def execute_application(
    app_module: JobbergateApplicationBase,
    app_config: JobbergateApplicationConfig,
    supplied_params: Optional[Dict[str, Any]] = None,
    fast_mode: bool = False,
):
    """
    Execute the jobbergate application python module.

    Updates the app_config with values gathered in the question workflow

    :param: app_module:      The source code for the application to execute
    :param: app_config:      The configuration for the JobbergateApplication
    :param: supplied_params: Pre-set values for the parameters. Any questions about these values will be skipped.
    :param: fast_mode:       If true, do not ask the user questions. Just use supplied_params or defaults
    :returns: The configuration values collected from the user by executing the application
    """
    app_params = gather_param_values(app_module, supplied_params=supplied_params, fast_mode=fast_mode)
    app_config.application_config.update(**app_params)
    return app_params



"""
Provide tool functions for working with Job Script data
"""

import json
import pathlib
from typing import Any, Dict, cast

from jobbergate_cli.exceptions import Abort
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import JobbergateContext, JobScriptResponse


def validate_parameter_file(parameter_path: pathlib.Path) -> Dict[str, Any]:
    """
    Validate parameter file at the supplied path and returns the parsed dict.

    Confirms:
        parameter_path exists
        parameter_path is a valid json file
    """
    data = None
    with Abort.check_expressions(
        f"The parameter file at {parameter_path} was invalid",
        raise_kwargs=dict(
            subject="Invalid parameter file",
            log_message=f"Parameter file located at {parameter_path} failed validation",
        ),
    ) as checker:
        checker(
            parameter_path.exists(),
            f"Parameter file {parameter_path} does not exist",
        )

        try:
            data = json.loads(parameter_path.read_text())
            is_valid_json = True
        except Exception:
            is_valid_json = False
        checker(is_valid_json, f"The parameter file at {parameter_path} is not valid JSON")

    # Make static type checkers happy
    assert data is not None
    return data


def fetch_job_script_data(
    jg_ctx: JobbergateContext,
    id: int,
) -> JobScriptResponse:
    """
    Retrieve a job_script from the API by ``id``
    """
    # Make static type checkers happy
    assert jg_ctx.client is not None

    return cast(
        JobScriptResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-scripts/{id}",
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve job script ({id}) from API",
            support=True,
            response_model_cls=JobScriptResponse,
        ),
    )



"""
Provide Pydantic models for various data items.
"""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

import httpx
import pydantic


class TokenSet(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    A model representing a pairing of access and refresh tokens
    """

    access_token: str
    refresh_token: Optional[str] = None


class IdentityData(pydantic.BaseModel):
    """
    A model representing the identifying data for a user from an auth token.
    """

    email: str
    client_id: str


class Persona(pydantic.BaseModel):
    """
    A model representing a pairing of a TokenSet and user email.
    This is a convenience to combine all of the identifying data and credentials for a given user.
    """

    token_set: TokenSet
    identity_data: IdentityData


class DeviceCodeData(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    A model representing the data that is returned from the OIDC provider's device code endpoint.
    """

    device_code: str
    verification_uri_complete: str
    interval: int


class JobbergateContext(pydantic.BaseModel, arbitrary_types_allowed=True):
    """
    A data object describing context passed from the main entry point.
    """

    persona: Optional[Persona]
    full_output: bool = False
    raw_output: bool = False
    client: Optional[httpx.Client]


class JobbergateConfig(pydantic.BaseModel):
    """
    A data object describing the config values needed in the "jobbergate_config" section of the
    JobbergateApplicationConfig model.
    """

    template_files: Optional[List[Path]]
    default_template: Optional[str] = None
    output_directory: Optional[Path] = None
    supporting_files_output_name: Optional[Dict[str, Any]] = None
    supporting_files: Optional[List[Path]] = None

    # For some reason, we support the application_config being about to override the *required*
    # job_script_name parameter that is passed at job_script creation time.
    # TODO: Find if this functionality is every used, and, if not, remove it immediately.
    job_script_name: Optional[str] = None


class JobbergateApplicationConfig(pydantic.BaseModel):
    """
    A data object describing the config data needed to instantiate a JobbergateApplication class.
    """

    jobbergate_config: JobbergateConfig
    application_config: Dict[str, Any]


class ApplicationResponse(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    Describes the format of data for applications retrieved from the Jobbergate API endpoints.
    """

    id: int
    application_name: str
    application_identifier: Optional[str] = None
    application_description: Optional[str] = None
    application_owner_email: str
    application_uploaded: bool
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    application_config: Optional[str] = None
    application_source_file: Optional[str] = None
    application_templates: Optional[Dict[str, str]] = None


class JobScriptFiles(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    Model containing job-script files.
    """

    main_file_path: Path
    files: Dict[str, str]


class JobScriptResponse(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    Describes the format of data for job_scripts retrieved from the Jobbergate API endpoints.
    """

    id: int
    application_id: int
    job_script_name: str
    job_script_description: Optional[str] = None
    job_script_files: JobScriptFiles
    job_script_owner_email: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None


class JobSubmissionResponse(pydantic.BaseModel, extra=pydantic.Extra.ignore):
    """
    Describes the format of data for job_submissions retrieved from the Jobbergate API endpoints.
    """

    id: int
    job_script_id: int
    cluster_name: Optional[str] = pydantic.Field(alias="client_id")
    slurm_job_id: Optional[int]
    execution_directory: Optional[Path]
    job_submission_name: str
    job_submission_description: Optional[str] = None
    job_submission_owner_email: str
    status: str
    created_at: Optional[datetime]
    updated_at: Optional[datetime]


class JobScriptCreateRequestData(pydantic.BaseModel):
    """
    Describes the data that will be sent to the ``create`` endpoint of the Jobbergate API for job scripts.
    """

    application_id: int
    job_script_name: str
    job_script_description: Optional[str]
    param_dict: Optional[JobbergateApplicationConfig] = None
    sbatch_params: Optional[List[Any]] = None


class JobSubmissionCreateRequestData(pydantic.BaseModel):
    """
    Describes the data that will be sent to the ``create`` endpoint of the Jobbergate API for job submissions.
    """

    job_submission_name: str
    job_submission_description: Optional[str] = None
    job_script_id: int
    client_id: Optional[str] = pydantic.Field(None, alias="cluster_name")
    execution_directory: Optional[Path] = None


class Pagination(pydantic.BaseModel):
    """
    A model describing the structure of the pagination component of a ListResponseEnvelope.
    """

    total: int
    start: Optional[int]
    limit: Optional[int]


class ListResponseEnvelope(pydantic.BaseModel):
    """
    A model describing the structure of response envelopes from "list" endpoints.
    """

    results: List[Dict[str, Any]]
    pagination: Pagination


class ClusterCacheData(pydantic.BaseModel):
    """
    Describes the format of data stored in the clusters cache file.
    """

    updated_at: datetime
    client_ids: List[str]


class ForeignKeyError(pydantic.BaseModel):
    """
    A model describing the structure of a foreign-key constraint error on delete.
    """

    message: str
    table: str
    pk_id: int



"""
Provide a stub module to maintain compatibility with previous versions.

Issue a deprecation warning when this module is imported from if JOBBERGATE_COMPATIBILITY_MODE is enabled.

If JOBBERGATE_COMPATIBILITY_MODE is not enabled, raise an import error when this module is imported.
"""

import warnings

from jobbergate_cli.config import settings
from jobbergate_cli.text_tools import dedent, unwrap


if settings.JOBBERGATE_COMPATIBILITY_MODE:

    from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase  # noqa

    warnings.warn(
        dedent(
            """
            Importing application_base from jobbergate_cli is deprecated.
            The module has been moved.
            Import 'application_base' from 'jobbergate_cli.subapps.applications' instead",
            """
        ),
        DeprecationWarning,
    )
else:
    raise ImportError(
        unwrap(
            """
            JobbergateApplicationBase has been moved to
            'jobbergate_cli.subapps.applications.application_base.JobbergateApplicationBase'
            """
        )
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jobbergate_cli',
 'jobbergate_cli.subapps',
 'jobbergate_cli.subapps.applications',
 'jobbergate_cli.subapps.clusters',
 'jobbergate_cli.subapps.job_scripts',
 'jobbergate_cli.subapps.job_submissions']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'boto3>=1.18.64,<2.0.0',
 'click>=8.0.1,<9.0.0',
 'httpx>=0.22.0,<0.23.0',
 'importlib-metadata>=4.2,<5.0',
 'inquirer>=2.7.0,<3.0.0',
 'loguru>=0.6.0,<0.7.0',
 'pendulum>=2.1.2,<3.0.0',
 'pep562>=1.1,<2.0',
 'py-buzz>=3.1.0,<4.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'python-dotenv>=0.19.0,<0.20.0',
 'python-jose>=3.3.0,<4.0.0',
 'rich>=11.2.0,<12.0.0',
 'sentry-sdk>=1.4.3,<2.0.0',
 'typer>=0.4.0,<0.5.0',
 'yarl>=1.7.2,<2.0.0']

entry_points = \
{'console_scripts': ['jobbergate = jobbergate_cli.main:app']}

setup_kwargs = {
    'name': 'jobbergate-cli',
    'version': '3.3.3a1',
    'description': 'Jobbergate CLI Client',
    'long_description': '================\n Jobbergate CLI\n================\n\nThe Jobbergate CLI provides a command-line interface to view and manage the Jobbergate\nresources. It can be used to create Job Scripts from template and then submit them to\nthe Slurm cluster to which Jobbergate is connected.\n\nJobbergate CLI is a Python project implemented with the\n`Typer <https://typer.tiangolo.com/>`_ CLI builder library. Its dependencies and\nenvironment are managed by `Poetry <https://python-poetry.org/>`_.\n\nThe CLI has a rich help system that can be accessed by passing the ``--help`` flag to\nthe main command:\n\n.. code-block:: console\n\n   jobbergate job-scripts --help\n\n\nThere is also help and parameter guides for each of the subcommands that can be accessed\nby passing them the ``--help`` flag:\n\n.. code-block:: console\n\n   jobbergate job-scripts list --help\n\nSee also:\n\n* `jobbergate-api <https://github.com/omnivector-solutions/jobbergate/jobbergate-api>`_\n\nLicense\n-------\n* `MIT <LICENSE>`_\n\n\nCopyright\n---------\n* Copyright (c) 2020 OmniVector Solutions <info@omnivector.solutions>\n',
    'author': 'Omnivector Solutions',
    'author_email': 'info@omnivector.solutions',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/omnivector-solutions/jobbergate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



"""
Provide initializers for logging.
"""

import sys

import sentry_sdk
from loguru import logger

from jobbergate_cli.config import settings


def init_logs(verbose=False):
    """
    Initialize logging.

    If JOBBERGATE_LOG_PATH is set in the config, add a rotatating file log handler.
    Logs will be retained for 1 week.

    If verbose is supplied, add a stdout handler at the DEBUG level.
    """
    # Remove default stderr handler at level INFO
    logger.remove()

    if verbose:
        logger.add(sys.stdout, level="DEBUG")

    if settings.JOBBERGATE_LOG_PATH is not None:
        logger.add(settings.JOBBERGATE_LOG_PATH, rotation="00:00", retention="1 week", level="DEBUG")
    logger.debug("Logging initialized")


def init_sentry():
    """
    Initialize Sentry if the ``SENTRY_DSN`` environment variable is present.
    """
    if settings.SENTRY_DSN:
        logger.debug("Initializing sentry")
        sentry_sdk.init(
            dsn=settings.SENTRY_DSN,
            traces_sample_rate=settings.SENTRY_TRACE_RATE,
            environment=settings.SENTRY_ENV,
        )
